''' Depth to PointCloud Tensorflow Subgraph
'''
from __future__ import print_function

import sys
import os
import tensorflow as tf
import numpy as np
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(BASE_DIR)
sys.path.append(BASE_DIR)
sys.path.append(os.path.join(ROOT_DIR, 'utils'))

import pdb

from scipy.spatial import Delaunay

H = 375
W = 1242
focal_len_pixels = 721.5377

# not sure whether this works as part of a TF graph...
def in_hull(p, hull):
    if not isinstance(hull,Delaunay):
        hull = Delaunay(hull)
    return hull.find_simplex(p)>=0

def box_2_ind(box2d, depthmap, npoints):
    #iterate over batch and create boolean mask for point selection
    box2d = box2d.numpy()
    depthmap = depthmap.numpy()
    npoints = npoints.numpy()
    indices = np.zeros((box2d.shape[0], npoints, 3), dtype=np.int32)
    for i in range(box2d.shape[0]):
        maskBox = np.zeros((box2d.shape[0], H, W), dtype=np.bool)
        xmin,ymin,xmax,ymax = np.ceil(box2d[i, :]).astype(np.int32)
        maskBox[i, ymin:ymax, xmin:xmax] = 1
        (y, x) = np.where(np.logical_and(\
                                np.logical_and(\
                                     (maskBox[i, :, :] == 1), depthmap[i, :, :] > 0), 
                                        depthmap[i, :, :] < 150 * focal_len_pixels))
        
        # workaround. find a better solution. 
        if (len(y) == 0):
            (y, x) = np.where(depthmap[i, :, :] > 0)

        np.random.seed(0)
        choice = np.random.choice(len(y), npoints, replace=True)
            
        y_sub = y[choice]
        x_sub = x[choice]

        indices[i, :, 0] = i
        indices[i, :, 1] = y_sub
        indices[i, :, 2] = x_sub

    return indices

def project_image_to_rect_tf(calib_mat, uv_depth):
    ''' Input: 
        calib: tensor of shape (B, 3, 4)
        uv_depth: (B, npoints, 3) array of u, v, and depth
        Output: Bxnx3 points in rect camera coord.
    '''
    calib_mat = calib_mat.numpy()

    batch_list = []
    
    for i in range(uv_depth.shape[0]):

        P = calib_mat[i, :, :]

        c_u = P[0,2]
        c_v = P[1,2]
        f_u = P[0,0]
        f_v = P[1,1]
        b_x = P[0,3]/(-f_u) # relative 
        b_y = P[1,3]/(-f_v)            

        z = uv_depth[i, :,2]
        x = tf.multiply((uv_depth[i,:,0]-c_u), z)/f_u + b_x
        y = tf.multiply((uv_depth[i,:,1]-c_v), z)/f_v + b_y
        
        xyz = tf.stack([x, y, z], axis = 1)
        batch_list.append(xyz)
    
    return tf.stack(batch_list, axis = 0)

# # rectify
# def project_image_to_rect(calib_mat, uv_depth):

#     calib_mat = calib_mat.numpy()
#     uv_depth = uv_depth.numpy()

#     pts_3d_rect = np.ones((uv_depth.shape[0], uv_depth.shape[1], 3), dtype=np.float32)

#     for i in range(uv_depth.shape[0]):

#         P = calib_mat[i, :, :]

#         c_u = P[0,2]
#         c_v = P[1,2]
#         f_u = P[0,0]
#         f_v = P[1,1]
#         b_x = P[0,3]/(-f_u) # relative 
#         b_y = P[1,3]/(-f_v)            

#         x = ((uv_depth[i,:,0]-c_u)*uv_depth[i,:,2])/f_u + b_x
#         y = ((uv_depth[i,:,1]-c_v)*uv_depth[i,:,2])/f_v + b_y
#         pts_3d_rect[i, :,0] = x
#         pts_3d_rect[i, :,1] = y
#         pts_3d_rect[i, :,2] = uv_depth[i, :,2]
        
#     return pts_3d_rect

def get_labels(pc_rect, box3d):
    pc_rect = pc_rect.numpy()
    box3d = box3d.numpy()
    labels = np.zeros((pc_rect.shape[0], pc_rect.shape[1]))

    for i in range(box3d.shape[0]):

        inds = in_hull(pc_rect[i, :, 0:3], box3d[i, :, :])

        labels[i, inds] = 1

    return labels


def rotate_pc_along_y_tf(pc, rot_angle):
    '''
    Input:
        pc: tensor (B, N,C), first 3 channels are XYZ
            z is facing forward, x is left ward, y is downward
        rot_angle: (B, 1) tensor
    Output: 
        pc: updated pc with XYZ rotated
    '''
    rot_angle = rot_angle.numpy()
    
    batch_list = []
    
    for i in range(pc.shape[0]):
        cosval = np.cos(rot_angle[i])
        sinval = np.sin(rot_angle[i])
        rotmat = np.array([[cosval, -sinval],[sinval, cosval]])
        rotmat = tf.convert_to_tensor(rotmat, tf.float32)
        
        # get x and z
        x = pc[i, :, 0]
        y = pc[i, :, 1]
        z = pc[i, :, 2]
        xz = tf.stack([x, z], axis = 1)
        
        # rotate
        xzt = tf.matmul(xz, tf.transpose(rotmat))
        x = xzt[:, 0]
        z = xzt[:, 1]
        
        # rebuild tensor and append result to list
        pc_rot = tf.stack([x, y, z], axis = 1)
        batch_list.append(pc_rot)
        
    return tf.stack(batch_list, axis = 0)


def depthmap_to_frustum_pointcloud_rot(depthmap, box2d, box3d, calib, rot_angle, npoints = 512):
    ''' Convert depthmap to a pointcloud within in frustum
    Input:
        depthmap: TF tensor in shape (B,W,H)
        box2d: TF tensor of shape (B, 4)
        calib: list of B calib objects
        npoints: int scalar, maximum number of points to keep (default: 512)
    Output:
        object_pc: TF tensor in shape (B,npoints,C) (C = 3 without intensity)
    '''
    # TODOS: 1) select points in frustum, 2) randomly subsample or resample points
    #        3) image to velo 4) velo to rect
    
    BATCH_SIZE = box2d.shape[0]
    
    npoints_tf = tf.constant(npoints, dtype = tf.int32)
    
    # get indices of points to retain npoints in bounding box
    indices = tf.py_function(box_2_ind, [box2d, depthmap, npoints_tf], tf.int32, name = 'box_2_ind')  
    indices = tf.reshape(indices, [BATCH_SIZE, npoints, 3])
    
    depth = tf.gather_nd(depthmap, indices)   # NOTE that shape: (B*npoints, )
    depth = tf.math.divide(depth, focal_len_pixels, name='pixel_to_meter')
    depth = tf.reshape(depth, [BATCH_SIZE, npoints]) # (B, npoints)
    indices_f = tf.cast(indices, tf.float32)
    uv_depth = tf.stack([indices_f[:, :, 2], indices_f[:, :, 1], depth], axis=2, name='stack_uv_depth')
    
    pc_rect = tf.py_function(project_image_to_rect_tf, [calib, uv_depth], tf.float32, name = 'uvd_to_rect')
    pc_rect = tf.reshape(pc_rect, [BATCH_SIZE, npoints, 3])
    
    labels = tf.py_function(get_labels, [pc_rect, box3d], tf.int32, name = 'get_labels')
    labels = tf.reshape(labels, [BATCH_SIZE, npoints])
    
    # rotate to center. important: so this after creating labels!    
    pc_rect = tf.py_function(rotate_pc_along_y_tf, [pc_rect, rot_angle], tf.float32, name = 'rot_to_center')
    pc_rect = tf.reshape(pc_rect, [BATCH_SIZE, npoints, 3])
        
    # add intensity channel
    intensity = tf.ones([BATCH_SIZE, npoints, 1])
    pc_rect = tf.concat([pc_rect, intensity], axis = 2)
    
    
    return pc_rect, labels